@@ -20,11 +20,13 @@
  */
 
 #include <uriparser/Uri.h>
-#include "CppTest/cpptest.h"
-#include "FourSuite.h"
+#include <cpptest.h>
 #include <memory>
 #include <stdio.h>
 #include <wchar.h>
+
+#include "FourSuite.h"
+
 using namespace Test;
 using namespace std;
 
@@ -58,6 +60,13 @@ public:
 		TEST_ADD(UriSuite::testUriUserInfoHostPort1)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort2)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort22_Bug1948038)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_1)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_2)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_3)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_4)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_related_1)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_related_12)
+		TEST_ADD(UriSuite::testUriUserInfoHostPort23_Bug3510198_related_2)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort3)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort4)
 		TEST_ADD(UriSuite::testUriUserInfoHostPort5)
@@ -68,6 +77,7 @@ public:
 		TEST_ADD(UriSuite::testUriHostIpSix1)
 		TEST_ADD(UriSuite::testUriHostIpSix2)
 		TEST_ADD(UriSuite::testUriHostIpFuture)
+		TEST_ADD(UriSuite::testUriHostEmpty)
 		TEST_ADD(UriSuite::testUriComponents)
 		TEST_ADD(UriSuite::testUriComponents_Bug20070701)
 		TEST_ADD(UriSuite::testEscaping)
@@ -84,6 +94,8 @@ public:
 		TEST_ADD(UriSuite::testNormalizeCrash_Bug20080224)
 		TEST_ADD(UriSuite::testFilenameUriConversion)
 		TEST_ADD(UriSuite::testCrash_FreeUriMembers_Bug20080116)
+		TEST_ADD(UriSuite::testCrash_Report2418192)
+		TEST_ADD(UriSuite::testPervertedQueryString);
 		TEST_ADD(UriSuite::testCrash_MakeOwner_Bug20080207)
 		TEST_ADD(UriSuite::testQueryList)
 		TEST_ADD(UriSuite::testQueryListPair)
@@ -457,7 +469,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		stateA.uri = &uriA;
 
 		int res;
-		
+
 		res = uriParseUriA(&stateA, "http://user:21@host/");
 		TEST_ASSERT(URI_SUCCESS == res);
 		TEST_ASSERT(!memcmp(uriA.userInfo.first, "user:21", 7 * sizeof(char)));
@@ -466,15 +478,155 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
 		TEST_ASSERT(uriA.portText.first == NULL);
 		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
 
 		res = uriParseUriA(&stateA, "http://user:1234@192.168.0.1:1234/foo.com");
 		TEST_ASSERT(URI_SUCCESS == res);
+		uriFreeUriMembersA(&uriA);
 
 		res = uriParseUriA(&stateA, "http://moo:21@moo:21@moo/");
 		TEST_ASSERT(URI_ERROR_SYNTAX == res);
+		uriFreeUriMembersA(&uriA);
 
 		res = uriParseUriA(&stateA, "http://moo:21@moo:21@moo:21/");
 		TEST_ASSERT(URI_ERROR_SYNTAX == res);
+		uriFreeUriMembersA(&uriA);
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_1() {
+		// User info with ":", with port, with escaped chars in password
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0         10 01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "user:%2F21" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "user:%2F21", 10 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 10);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_2() {
+		// User info with ":", with port, with escaped chars in user name and password
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0            13 01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "%2Fuser:%2F21" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "%2Fuser:%2F21", 13 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 13);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_3() {
+		// User info with ":", with port, with escaped chars in password
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0               16 01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "user:!$&'()*+,;=" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "user:!$&'()*+,;=", 16 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 16);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_4() {
+		// User info with ":", with port, with escaped chars in user name and password
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0                   20 01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "!$&'()*+,;=:password" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "!$&'()*+,;=:password", 20 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 20);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_related_1() {
+		// Empty user info
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(uriA.userInfo.afterLast != NULL);
+		TEST_ASSERT(uriA.userInfo.first != NULL);
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 0);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_related_12() {
+		// Empty user info
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0      7  01
+		res = uriParseUriA(&stateA, "http" "://" "%2Fhost" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(uriA.userInfo.afterLast == NULL);
+		TEST_ASSERT(uriA.userInfo.first == NULL);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "%2Fhost", 7 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 7);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
+	}
+
+	void testUriUserInfoHostPort23_Bug3510198_related_2() {
+		// Several colons in userinfo
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+
+		int res;
+		//                           0   4  0  3  0 2  01  0   4  01
+		res = uriParseUriA(&stateA, "http" "://" "::" "@" "host" "/");
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(!memcmp(uriA.userInfo.first, "::", 2 * sizeof(char)));
+		TEST_ASSERT(uriA.userInfo.afterLast - uriA.userInfo.first == 2);
+		TEST_ASSERT(!memcmp(uriA.hostText.first, "host", 4 * sizeof(char)));
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 4);
+		TEST_ASSERT(uriA.portText.first == NULL);
+		TEST_ASSERT(uriA.portText.afterLast == NULL);
+		uriFreeUriMembersA(&uriA);
 	}
 
 	void testUriUserInfoHostPort3() {
@@ -636,6 +788,24 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		uriFreeUriMembersA(&uriA);
 	}
 
+	void testUriHostEmpty() {
+		UriParserStateA stateA;
+		UriUriA uriA;
+		stateA.uri = &uriA;
+		//                          0   4  0  3  01  0  3
+		const char * const input = "http" "://" ":" "123";
+		const int res = uriParseUriA(&stateA, input);
+		TEST_ASSERT(URI_SUCCESS == res);
+		TEST_ASSERT(uriA.userInfo.first == NULL);
+		TEST_ASSERT(uriA.userInfo.afterLast == NULL);
+		TEST_ASSERT(uriA.hostText.first != NULL);
+		TEST_ASSERT(uriA.hostText.afterLast != NULL);
+		TEST_ASSERT(uriA.hostText.afterLast - uriA.hostText.first == 0);
+		TEST_ASSERT(uriA.portText.first == input + 4 + 3 + 1);
+		TEST_ASSERT(uriA.portText.afterLast == input + 4 + 3 + 1 + 3);
+		uriFreeUriMembersA(&uriA);
+	}
+
 	void testUriHostIpFuture() {
 		// TODO
 	}
@@ -702,12 +872,14 @@ Rule                                | Example | hostSet | absPath | emptySeg
 
 	bool testUnescapingHelper(const wchar_t * input, const wchar_t * output,
 			bool plusToSpace = false, UriBreakConversion breakConversion = URI_BR_DONT_TOUCH) {
-		wchar_t * working = static_cast<URI_CHAR *>(malloc((URI_STRLEN(input) + 1) * sizeof(URI_CHAR)));
+		wchar_t * working = new wchar_t[URI_STRLEN(input) + 1];
 		wcscpy(working, input);
 		const wchar_t * newTermZero = uriUnescapeInPlaceExW(working,
 				plusToSpace ? URI_TRUE : URI_FALSE, breakConversion);
-		return ((newTermZero == working + wcslen(output))
+		const bool success = ((newTermZero == working + wcslen(output))
 				&& !wcscmp(working, output));
+		delete[] working;
+		return success;
 	}
 
 	void testUnescaping() {
@@ -806,6 +978,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		stateW.uri = &baseUri;
 		int res = uriParseUriW(&stateW, base);
 		if (res != 0) {
+			uriFreeUriMembersW(&baseUri);
 			return false;
 		}
 
@@ -815,6 +988,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		res = uriParseUriW(&stateW, rel);
 		if (res != 0) {
 			uriFreeUriMembersW(&baseUri);
+			uriFreeUriMembersW(&relUri);
 			return false;
 		}
 
@@ -825,6 +999,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		if (res != 0) {
 			uriFreeUriMembersW(&baseUri);
 			uriFreeUriMembersW(&relUri);
+			uriFreeUriMembersW(&expectedUri);
 			return false;
 		}
 
@@ -834,6 +1009,8 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		if (res != 0) {
 			uriFreeUriMembersW(&baseUri);
 			uriFreeUriMembersW(&relUri);
+			uriFreeUriMembersW(&expectedUri);
+			uriFreeUriMembersW(&transformedUri);
 			return false;
 		}
 
@@ -851,6 +1028,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		uriFreeUriMembersW(&baseUri);
 		uriFreeUriMembersW(&relUri);
 		uriFreeUriMembersW(&expectedUri);
+		uriFreeUriMembersW(&transformedUri);
 		return equal;
 	}
 
@@ -1043,7 +1221,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(testToStringHelper(uriTwoString));
 	}
 
-	bool testToStringCharsRequiredHelper(wchar_t * text) {
+	bool testToStringCharsRequiredHelper(const wchar_t * text) {
 		// Parse
 		UriParserStateW state;
 		UriUriW uri;
@@ -1093,7 +1271,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(testToStringCharsRequiredHelper(L"test"));
 	}
 
-	bool testNormalizeMaskHelper(wchar_t * uriText, unsigned int expectedMask) {
+	bool testNormalizeMaskHelper(const wchar_t * uriText, unsigned int expectedMask) {
 		UriParserStateW state;
 		UriUriW uri;
 		state.uri = &uri;
@@ -1134,7 +1312,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(testNormalizeMaskHelper(L"http://localhost/#AB%43", URI_NORMALIZE_FRAGMENT));
 	}
 
-	bool testNormalizeSyntaxHelper(wchar_t * uriText, wchar_t * expectedNormalized,
+	bool testNormalizeSyntaxHelper(const wchar_t * uriText, const wchar_t * expectedNormalized,
 			unsigned int mask = static_cast<unsigned int>(-1)) {
 		UriParserStateW stateW;
 		int res;
@@ -1359,6 +1537,34 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(true);
 	}
 
+	void testCrash_Report2418192() {
+		// Testcase by Harvey Vrsalovic
+		helperTestQueryString("http://svcs.cnn.com/weather/wrapper.jsp?&csiID=csi1", 1);
+	}
+
+	void testPervertedQueryString() {
+		helperTestQueryString("http://example.org/?&&=&&&=&&&&==&===&====", 5);
+	}
+
+	void helperTestQueryString(char const * uriString, int pairsExpected) {
+		UriParserStateA state;
+		UriUriA uri;
+		state.uri = &uri;
+		int res = uriParseUriA(&state, uriString);
+		TEST_ASSERT(res == URI_SUCCESS);
+
+		UriQueryListA * queryList = NULL;
+		int itemCount = 0;
+
+		res = uriDissectQueryMallocA(&queryList, &itemCount,
+				uri.query.first, uri.query.afterLast);
+		TEST_ASSERT(res == URI_SUCCESS);
+		TEST_ASSERT(queryList != NULL);
+		TEST_ASSERT(itemCount == pairsExpected);
+		uriFreeQueryListA(queryList);
+		uriFreeUriMembersA(&uri);
+	}
+
 	void testCrash_MakeOwner_Bug20080207() {
 		// Testcase by Adrian Manrique
 		UriParserStateA state;
@@ -1377,7 +1583,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 
 	void testQueryListHelper(const wchar_t * input, int expectedItemCount) {
 		int res;
-		
+
 		UriBool spacePlusConversion = URI_TRUE;
 		UriBool normalizeBreaks = URI_FALSE;
 		UriBreakConversion breakConversion = URI_BR_DONT_TOUCH;
@@ -1391,9 +1597,9 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT((queryList == NULL) == (expectedItemCount == 0));
 
 		if (expectedItemCount != 0) {
-			// First 
+			// First
 			int charsRequired;
-			res = uriComposeQueryCharsRequiredExW(queryList, &charsRequired, spacePlusConversion, 
+			res = uriComposeQueryCharsRequiredExW(queryList, &charsRequired, spacePlusConversion,
 					normalizeBreaks);
 			TEST_ASSERT(res == URI_SUCCESS);
 			TEST_ASSERT(charsRequired >= (int)wcslen(input));
@@ -1434,7 +1640,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		int res;
 		UriQueryListA * queryList;
 		int itemCount;
-		
+
 		res = uriDissectQueryMallocA(&queryList, &itemCount, pair, pair + strlen(pair));
 		TEST_ASSERT(res == URI_SUCCESS);
 		TEST_ASSERT(queryList != NULL);
@@ -1448,6 +1654,7 @@ Rule                                | Example | hostSet | absPath | emptySeg
 		TEST_ASSERT(recomposed != NULL);
 		TEST_ASSERT(!strcmp(recomposed, (fixed != NULL) ? fixed : pair));
 		free(recomposed);
+		uriFreeQueryListA(queryList);
 	}
 
 	void testQueryListPair() {
